@@ -6,10 +6,10 @@ void *mh_xrealloc(void *, size_t);
 void *mh_xcalloc(size_t, size_t);
 void mh_free0(void *);
 char *pwd(void);
-char *add(char *, char *);
+char *add(const char *, char *);
 void create_folder(char *, int, void (*)(int));
 int num_digits(int);
-char *mh_xstrdup(char *);
+char *mh_xstrdup(const char *);
 
 struct msgs_array {
 	int max, size;
@@ -17,3 +17,4 @@ struct msgs_array {
 };
 
 void app_msgarg(struct msgs_array *, char *);
+int parse_msgs(const struct msgs_array *, char *, struct msgs_array *);
