@@ -29,11 +29,7 @@ ext_hook(char *hook_name, char *msg_filename1, char *msg_filename2)
 		if (did_message) {
 			return (NOTOK);
 		}
-		if ((hook = context_find("msg-hook"))) {
-			advise(NULL, hook);
-		} else {
-			advise(NULL, "external hook (%s) failed.", hook);
-		}
+		advise(NULL, "external hook (%s) failed.", hook);
 		did_message = 1;
 		return (NOTOK);
 	} else {
