@@ -31,3 +31,15 @@ trim(unsigned char *cp)
 
 	return cp;
 }
+
+char *
+rtrim(char *cp)
+{
+	char *sp = cp + strlen(cp) - 1;
+
+	while (sp >= cp && isspace(*sp)) {
+		sp--;
+	}
+	*++sp = '\0';
+	return cp;
+}
