@@ -234,6 +234,7 @@ get_content(FILE *in, char *file, int toplevel)
 	enum state state;
 	struct field f = {{0}};
 	int compnum;
+	char *buf;
 	CT ct;
 	HF hp;
 
@@ -254,10 +255,22 @@ get_content(FILE *in, char *file, int toplevel)
 			state = FLD2;
 			/* FALL */
 		case FLD2:
+			if (compnum == 1) {
+				ct->crlf = f.value[f.valuelen-2] == '\r';
+			}
 			compnum++;
 
+			/* decode rfc2047 */
+			buf = mh_xcalloc(sizeof(char *), f.valuelen);
+			if (!decode_rfc2047(f.value, buf, f.valuelen)) {
+				mh_free0(&buf);
+				buf = mh_xstrdup(f.value);
+			}
+
 			/* add the header data to the list */
-			add_header(ct, mh_xstrdup(f.name), mh_xstrdup(f.value));
+			add_header(ct, mh_xstrdup(f.name), buf);
+
+			buf = NULL;
 
 			ct->c_begin = ftell(in) + 1;
 			continue;
@@ -550,13 +563,19 @@ incl_name_value(unsigned char *buf, char *name, char *value) {
 ** one, return the entire value.  Note that, for example, a name_suffix
 ** of name will match filename="foo", and return foo.
 */
-static char *
+char *
 extract_name_value(char *name_suffix, char *value) {
-	char *extracted_name_value = value;
-	char *name_suffix_plus_quote = concat(name_suffix, "=\"", NULL);
-	char *name_suffix_equals = strstr(value, name_suffix_plus_quote);
+	char *extracted_name_value;
+	char *name_suffix_plus_quote;
+	char *name_suffix_equals;
 	char *cp;
 
+	if (!value) {
+		return value;
+	}
+	extracted_name_value = value;
+	name_suffix_plus_quote = concat(name_suffix, "=\"", NULL);
+	name_suffix_equals = strstr(value, name_suffix_plus_quote);
 	mh_free0(&name_suffix_plus_quote);
 	if (name_suffix_equals) {
 		char *name_suffix_begin;
@@ -1019,7 +1038,7 @@ InitMultiPart(CT ct)
 	*/
 	if (ct->c_encoding != CE_7BIT && ct->c_encoding != CE_8BIT
 		&& ct->c_encoding != CE_BINARY) {
-		admonish(NULL, "\"%s/%s\" type in message %s must be encoded in 7bit, 8bit, or binary", ci->ci_type, ci->ci_subtype, ct->c_file);
+		admonish(NULL, "\"%s/%s\" type in message %s should be encoded in 7bit, 8bit, or binary", ci->ci_type, ci->ci_subtype, ct->c_file);
 		ct->c_encoding = CE_7BIT;
 	}
 
@@ -1066,8 +1085,14 @@ InitMultiPart(CT ct)
 	*++dp = '\0';
 
 	/* record boundary separators */
-	m->mp_start = concat(bp, "\n", NULL);
-	m->mp_stop = concat(bp, "--\n", NULL);
+	if (!ct->crlf) {
+		m->mp_start = concat(bp, "\n", NULL);
+		m->mp_stop = concat(bp, "--\n", NULL);
+	} else {
+		m->mp_start = concat(bp, "\r\n", NULL);
+		m->mp_stop = concat(bp, "--\r\n", NULL);
+	}
+
 
 	if (!ct->c_fp && (ct->c_fp = fopen(ct->c_file, "r")) == NULL) {
 		advise(ct->c_file, "unable to open for reading");
@@ -1237,9 +1262,9 @@ InitMessage(CT ct)
 	struct k2v *kv;
 	CI ci = &ct->c_ctinfo;
 
-	if ((ct->c_encoding != CE_7BIT) && (ct->c_encoding != CE_8BIT)) {
-		admonish(NULL, "\"%s/%s\" type in message %s should be encoded in 7bit or 8bit", ci->ci_type, ci->ci_subtype, ct->c_file);
-		return NOTOK;
+	if ((ct->c_encoding != CE_7BIT) && (ct->c_encoding != CE_8BIT) && (ct->c_encoding != CE_BINARY)) {
+		admonish(NULL, "\"%s/%s\" type in message %s should be encoded in 7bit, 8bit, or binary", ci->ci_type, ci->ci_subtype, ct->c_file);
+		ct->c_encoding = CE_7BIT;
 	}
 
 	/* check for missing subtype */
